"""
A script that checks an existing ITSI environment for potentially misconfigured entities used in KPIs.

Example usage:

    $ $SPLUNK_HOME/bin/splunk cmd python check_kpi_entity_configs.py

"""

import getpass
import json
import os
import sys

from splunk import auth
from splunk.rest import simpleRequest


SPLUNKD_HOST = os.environ.get('SPLUNKD_HOST', 'https://localhost')
SPLUNKD_PORT = os.environ.get('SPLUNKD_PORT', '8089')
SPLUNKWEB_PORT = os.environ.get('SPLUNKWEB_PORT', '8000')
SPLUNK_USER = os.environ.get('SPLUNK_USER', 'admin')

LOG_TAG = 'EntityMisconfig'

MISSING_FILTER_FIELD = (
    'entity_title="{entity_title}" (entity_id="{entity_id}") is missing an alias for the entity_filter_field="{field}" of kpi_title="{kpi_title}" (kpi_id="{kpi_id}"") in service_title="{service_title}" (service_id="{service_id}"). '
    'The entity might not be included in this KPI\'s results.'
)

MISSING_SPLIT_FIELD = (
    'entity_title="{entity_title}" (entity_id="{entity_id}") is missing an alias for the entity_split_field="{field}" of kpi_title="{kpi_title}" (kpi_id="{kpi_id}") in service_title="{service_title}" (service_id="{service_id}"). '
    'The entity will not be included in this KPI\'s results.'
)

MISSING_ALIAS_FILTERING_VALUES = (
    'entity_title="{entity_title}" (entity_id="{entity_id}") has potentially missing_alias_values="{missing_alias_values}" for the entity_filter_field="{field}" of kpi_title="{kpi_title}" (kpi_id="{kpi_id}") in service_title="{service_title}" (service_id="{service_id}"). '
    'The entity might not be included in this KPI\'s results due to the removal of the previous KPI entity_alias_filtering_fields="{alias_filtering_fields}".'
)

def login_to_splunk():
    splunkd_host = raw_input('Enter splunkd host. Press enter to use {}: '.format(SPLUNKD_HOST))
    if not splunkd_host.strip():
        splunkd_host = SPLUNKD_HOST

    splunkd_port = None
    retries = 3

    while not splunkd_port and retries:
        splunkd_port = raw_input('Enter splunkd port. Press enter to use {}: '.format(SPLUNKD_PORT))

        if not splunkd_port.strip():
            splunkd_port = SPLUNKD_PORT
        else:
            try:
                int(splunkd_port)
            except ValueError:
                retries -= 1
                print 'Invalid port. Try again. {} tries left'.format(retries)
                continue
    if not retries and not splunkd_port:
        print 'You have reached the maximum number of retries. Run me again.'
        sys.exit(1)

    hostpath = '{}:{}'.format(splunkd_host, splunkd_port).strip()
    print 'Your Splunk instance is: {}'.format(hostpath)

    session_key = None
    while not session_key and retries:
        username = raw_input('Enter the Splunk user. Press enter to use {}: '.format(SPLUNK_USER))
        if not username.strip():
            username = SPLUNK_USER

        password = getpass.getpass(prompt='Enter password for {}: '.format(username))

        print 'Trying to obtain a Splunk session key...'
        try:
            session_key = auth.getSessionKey(username, password, hostpath)
        except Exception as e:
            retries -= 1
            print 'Encountered an error while trying to log you in - {}'.format(e)
            print 'Let\'s try again... {} retries left'.format(retries)

    return session_key, splunkd_host, splunkd_port


def check_entity_misconfig(session_key):
    """
    Checks all KPIs and their entities for potential misconfigurations.

    @return: True if all checks pass, False otherwise
    @rtype: bool
    """
    services_dict = construct_service_to_entities_dict(session_key)
    if not services_dict:
        return True

    failed = False

    services = fetch_objects(session_key, 'service')

    for service in services:
        kpis = service.get('kpis')
        if not kpis:
            continue

        service_key = service.get('_key')
        entities_in_service = services_dict.get(service_key, [])

        for kpi in kpis:
            passed_config_check = check_kpi_entity_config(service, kpi, entities_in_service)
            failed = failed or (not passed_config_check)

    return not failed


def fetch_objects(session_key, object_type):
    """
    Fetches ITOA objects data based on the supplied object type.

    @param session_key: the session key
    @type session_key: basestring
    @param object_name: the object name
    @type object_name: basestring
    @return: a list of data objects
    """
    url = '{}:{}/servicesNS/nobody/SA-ITOA/itoa_interface/{}?count=0'.format(
        SPLUNKD_HOST,
        SPLUNKD_PORT,
        object_type
    )
    response, content = simpleRequest(url, session_key, timeout=300)
    return json.loads(content)


def construct_service_to_entities_dict(session_key):
    """
    Constructs a lookup dictionary based on service keys, and their associated entities.

    @return: a dictionary from service key to their associated entities
    @rtype: dict
    """
    services_dict = {}

    entities = fetch_objects(session_key, 'entity')
    for entity in entities:
        services = entity.get('services', [])
        for service in services:
            service_key = service.get('_key')
            service_entities = services_dict.setdefault(service_key, [])
            service_entities.append(entity)

    return services_dict


def check_kpi_entity_config(service, kpi, entities):
    """
    Checks the given KPI against the service's entities for a valid configuration.

    @param service: the service data
    @type service: dict
    @param kpi: the kpi data
    @type kpi: dict
    @param entities: the entities data
    @type entities: list of dict
    @return: True if the check passes, False otherwise
    @rtype: bool
    """
    failed = False

    for entity in entities:
        split_field = kpi.get('entity_breakdown_id_fields')
        filter_field = kpi.get('entity_id_fields')
        alias_filtering_fields = kpi.get('entity_alias_filtering_fields')

        passed_split_field_check = check_split_field(kpi, entity)
        if not passed_split_field_check:
            log_field_misconfiguration(MISSING_SPLIT_FIELD, split_field, service, kpi, entity)

        passed_filter_field_check = check_filter_field(kpi, entity)
        if not passed_filter_field_check:
            log_field_misconfiguration(MISSING_FILTER_FIELD, filter_field, service, kpi, entity)

        missing_alias_values = set()
        passed_alias_filtering_check = check_alias_filtering_field(kpi, entity, missing_alias_values)
        if not passed_alias_filtering_check:
            alias_values = ', '.join(missing_alias_values)
            log_field_misconfiguration(
                MISSING_ALIAS_FILTERING_VALUES,
                filter_field,
                service,
                kpi,
                entity,
                alias_filtering_fields=alias_filtering_fields or '',
                missing_alias_values=alias_values
            )

        failed = failed or (not passed_split_field_check) or (not passed_filter_field_check) or (not passed_alias_filtering_check)

    return not failed


def check_split_field(kpi, entity):
    """
    Checks whether the KPI's Split Field setting is misconfigured for the given entity.

    @param kpi: the kpi data
    @type kpi: dict
    @param entity: the entity data
    @type entity: dict
    @return: True if the check passes, False otherwise
    @rtype: bool
    """
    is_entity_breakdown = kpi.get('is_entity_breakdown', False)
    if not is_entity_breakdown:
        return True

    split_field = kpi.get('entity_breakdown_id_fields')
    return check_entity_for_alias(entity, split_field)


def check_filter_field(kpi, entity):
    """
    Checks whether the KPI's Filter Field setting is misconfigured for the given entity.

    @param kpi: the kpi data
    @type kpi: dict
    @param entity: the entity data
    @type entity: dict
    @return: True if the check passes, False otherwise
    @rtype: bool
    """
    is_service_entity_filter = kpi.get('is_service_entity_filter', False)
    if not is_service_entity_filter:
        return True

    filter_field = kpi.get('entity_id_fields')
    return check_entity_for_alias(entity, filter_field)


def check_alias_filtering_field(kpi, entity, missing_alias_values):
    """
    Checks whether the KPI's previous Entity Alias Filtering fields may cause a misconfiguration for the given entity.

    @param kpi: the kpi data
    @type kpi: dict
    @param entity: the entity data
    @type entity: dict
    @param missing_alias_values: a set that's populated with any missing alias values
    @type missing_alias_values: set
    @return: True if the check passes, False otherwise
    @rtype: bool
    """
    is_service_entity_filter = kpi.get('is_service_entity_filter', False)
    if not is_service_entity_filter:
        return True

    alias_filtering = kpi.get('entity_alias_filtering_fields')
    if not alias_filtering:
        # If Entity Alias Filtering is blank, previously all alias fields were used
        alias_filtering_fields = entity.get('identifier', {}).get('fields', [])
    elif isinstance(alias_filtering, basestring):
        alias_filtering_fields = alias_filtering.split(',')
    elif not isinstance(alias_filtering, list):
        # Unexpected type, pass the check
        return True

    filter_field = kpi.get('entity_id_fields')
    filter_values = set(entity.get(filter_field, []))

    # Check that the other Entity Alias Filtering field values are present in the Entity Filter Field for the entity
    for alias_filter in alias_filtering_fields:
        alias_filter_values = set(entity.get(alias_filter.strip(), []))
        missing_values = alias_filter_values - filter_values
        if missing_values:
            # Add any potentially missing alias values in the Entity Filter Field
            missing_alias_values.update(missing_values)

    if missing_alias_values:
        # If there are any potentially missing alias values, fail the check
        return False

    return True


def check_entity_for_alias(entity, alias_field):
    """
    Checks whether the entity has the given alias field.

    @param entity: the entity data
    @type entity: dict
    @param alias_field: the alias field name
    @type alias_field: str
    @returns: True if the entity has the alias, False otherwise
    @rtype: bool
    """
    identifiers = entity.get('identifier', {}).get('fields', [])
    return alias_field in entity and alias_field in identifiers


def log_field_misconfiguration(base_message, field, service, kpi, entity, **kwargs):
    """
    Logs a message detailing a potential misconfiguration in KPI and Entity setup.

    @param base_message: the base message string
    @type base_message: str
    @param field: the KPI config field
    @type field: str
    @param service: the service data
    @type service: dict
    @param kpi: the kpi data
    @type kpi: dict
    @param entity: the entity data
    @type entity: dict
    """
    formatted = base_message.format(
        entity_id=entity.get('_key'),
        entity_title=entity.get('title'),
        field=field,
        kpi_id=kpi.get('_key'),
        kpi_title=kpi.get('title'),
        service_id=service.get('_key'),
        service_title=service.get('title'),
        **kwargs
    )
    message = '<{}> {}'.format(LOG_TAG, formatted)
    print message


if __name__ == '__main__':
    session_key, splunkd_host, splunkd_port = login_to_splunk()

    SPLUNKD_HOST = splunkd_host
    SPLUNKD_PORT = splunkd_port

    print 'Checking KPI and their entity configurations...'
    print

    check_entity_misconfig(session_key)

    print
    print 'Done.'
